//  ____                    _
// | __ )  ___   __ _  __ _| | ___   ___ _ __  _ __
// |  _ \ / _ \ / _` |/ _` | |/ _ \ / __| '_ \| '_ \
// | |_) | (_) | (_| | (_| | |  __/| (__| |_) | |_) |
// |____/ \___/ \__, |\__, |_|\___(_)___| .__/| .__/
//              |___/ |___/             |_|   |_|
// Author: Love Billenius <lovbi127@student.liu.se>

#include "Boggle.h"
#include "shuffle.h"
#include <unordered_set>

// The res folder isn't copied the same way on macOS
#ifdef __APPLE__
const std::string Boggle::s_DICTIONARY_FILE = "../../../EnglishWords.dat";
#else
const std::string Boggle::DICTIONARY_FILE = "EnglishWords.dat";
#endif

static const int g_NUM_CUBES = 16; // the number of cubes in the game
static const int g_CUBE_SIDES = 6; // the number of sides on each cube
static std::string g_CUBES[g_NUM_CUBES] =
    { // the letters on all 6 sides of every cube
    "AAEEGN", "ABBJOO", "ACHOPS", "AFFKPS", "AOOTTW", "CIMOTU",
    "DEILRX", "DELRVY", "DISTTY", "EEGHNW", "EEINSU", "EHRTVW",
    "EIOSST", "ELRTTY", "HIMNQU", "HLNNRZ"};

std::array<std::string, g_NUM_ROLLED_CUBES> rollSides() {
    // Intermediate array to shuffle
    std::array<std::string, g_NUM_CUBES> cubes;
    std::copy(std::begin(g_CUBES), std::end(g_CUBES), cubes.begin());
    shuffle(cubes.data(), cubes.size());

    // Just return the first g_NUM_ROLLED_CUBES (4pcs) entries
    std::array<std::string, g_NUM_ROLLED_CUBES> ret;
    std::copy(cubes.begin(), cubes.begin() + g_NUM_ROLLED_CUBES, ret.begin());
    return ret;
}

Boggle::Boggle()
    : m_showingSides(rollSides()), m_englishWords(Boggle::s_DICTIONARY_FILE) {
    std::cout << "Lexicon size: " << m_englishWords.size() << std::endl;
}

const std::vector<string> &Boggle::playedWords() const { return m_playedWords; }

const std::array<string, g_NUM_ROLLED_CUBES> Boggle::showingSides() const {
    return m_showingSides;
}

bool Boggle::isWordEnglish(const std::string &word) const {
    return m_englishWords.contains(word);
}

bool Boggle::isWordLongEnough(const std::string &word) const {
    return word.length() >= s_MIN_WORD_LENGTH;
}

bool Boggle::isWordPlayed(const std::string &word) const {
    return std::find(m_playedWords.begin(), m_playedWords.end(), word) !=
           m_playedWords.end();
}

unsigned int Boggle::getPointsForWord(const std::string &word) {
    // This is kept as a variable to avoid static_cast<int>
    const int pointsForValidWord = word.length() - s_MIN_WORD_LENGTH;
    return std::max(pointsForValidWord, 0);
}

bool Boggle::userInsert(std::string word) {
    bool existsAlready = std::find(m_playedWords.begin(), m_playedWords.end(),
                                   word) != m_playedWords.end();
    if (existsAlready)
        return false;

    m_playedWords.push_back(std::move(word));
    return true;
}

std::vector<std::string> Boggle::computerPlay() const {
    std::unordered_set<std::string> visited;
    std::vector<std::string> valiWords;

    std::cout << "showing sides: "<<m_showingSides.size() <<std::endl;
    for (const std::string &cube : m_showingSides) {
        visited.insert(cube);
        std::string word;
        for (const char c : cube) {
            word.push_back(c);
            backtrack(valiWords, visited, word);
        }
    }

    return valiWords;
}

void Boggle::backtrack(std::vector<std::string> &validWords,
                       std::unordered_set<string> &visited,
                       std::string &word) const {
    if (!m_englishWords.containsPrefix(word))
        return;

    if (word.size() >= s_MIN_WORD_LENGTH && m_englishWords.contains(word))
        validWords.push_back(word);

    for (const std::string &neighbor : m_showingSides) {
        bool neighborAlreadyVisited = visited.find(neighbor) != visited.end();
        if (neighborAlreadyVisited)
            continue;

        visited.insert(word);
        for (const char neighborChar : neighbor) {
            word.push_back(neighborChar);
            backtrack(validWords, visited, word);
            word.pop_back();
        }
    }
}
